/*
   This file is part of a program called sphereEversion.
   The complete source code can be downloaded from
      http://www.dgp.toronto.edu/~mjmcguff/eversion/

   This file was written by Michael McGuffin, 1998,
   and has undergone slight modifications since then.
*/
#include <stdio.h>    /* sprintf(), printf(), fprintf(), getchar() */
#include <string.h>   /* memcpy(), strcat() */
#include <stdlib.h>   /* exit() on Linux */
#include <unistd.h>   /* exit() on Irix */
#include <libgen.h>   /* basename() */
#include <math.h>     /* acos() */

/* #include <X11/X.h> */
#include <X11/Xlib.h>
/* #include <X11/Xutil.h> */
#include <X11/keysym.h>

#include "vroot.h"     /* very important! */


#include <GL/gl.h>
#include <GL/glx.h>   /* glX*() routines */
/* #include <GL/glu.h> */

/* #define TOGGLEABLE_WINDOW_DECORATIONS */  /* requires Motif libraries */

#ifdef TOGGLEABLE_WINDOW_DECORATIONS
#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#endif

#include "generateGeometry.h"

/* typedef enum { False = 0, True = 1 } Boolean; */
#ifndef TOGGLEABLE_WINDOW_DECORATIONS
typedef unsigned char Boolean;
#endif
typedef unsigned long Pixel;

const double PI = 3.14159265358979323846264338327950;

const float objectRadius = 1.5;
const float objectDistanceFromCamera = 1.0;

const double speedOfAnimatedRotation = 0.01; // must be between 0.0 and 1.0

// ==========================================================================

// global variables

double deltaTime = 1.0/64;

struct {
   char * commandLineName;

   // Recall that in X terminology,
   // one *display* can consist of
   // many *screens*.
   //
   Display *display;
   int screen_num;
   Screen *screen_ptr;
   int screenWidth, screenHeight;      // screen dimensions in pixels
   int screenWidthMM, screenHeightMM;  // screen dimensions in millimeters
   unsigned int screenDepth;
   Window rootWindow;
   Window window;
   unsigned int windowWidth, windowHeight;   /* dimensions in pixels */
   unsigned int minOfWidthAndHeight() {
      return windowWidth < windowHeight ? windowWidth : windowHeight;
   }
   Cursor cursor;
   Pixel blackPixel, whitePixel;
   GC gc;   /* Graphics context. */
   XFontStruct *bigFont, *smallFont;
   Colormap colourMap;
} XStuff;

// display parameters
const int defaultNumberOfLatitudinalPatchesPerHemisphere = 12;
const int defaultNumberOfLongitudinalPatchesPerStrip = 12;
const int defaultNumStrips = 8;
Boolean bBackFaceCulling = False;
Boolean bFrontFacing = True;
Boolean bShadingIsSmooth = True;
typedef enum {
   style_points = 0,
   style_polygons,
   style_checkered,
   style_bands,
   number_of_styles
} RenderingStyle;
RenderingStyle renderingStyle = style_polygons;
Boolean bHalfStrips = False;
Boolean bAnimatedEversion = False;
Boolean bAnimatedEversionBackwards = False;
Boolean bAnimatedRotation = False;
Boolean bDisplayText = True;
Boolean bRootMode = False;
#ifdef TOGGLEABLE_WINDOW_DECORATIONS
Boolean bNoWindowDecorations = False;
#endif
double angleOfMouseDragForAnimatedRotation = 0.0; // radians
Pixel currentTextColour;

// ==========================================================================

Pixel AllocPixelColour(double r, double g, double b) {

   XColor xcolor;

   xcolor.pixel = 0;
   xcolor.red   = (int)(r * 65535 + 0.5);  // add .5 to round-off to nearest integer
   xcolor.green = (int)(g * 65535 + 0.5);
   xcolor.blue  = (int)(b * 65535 + 0.5);
   xcolor.flags = DoRed | DoGreen | DoBlue;
   XAllocColor(XStuff.display, XStuff.colourMap, &xcolor);
   return xcolor.pixel;
}

void FreePixelColour(Pixel p) {

   XFreeColors(XStuff.display, XStuff.colourMap, &p, 1, 0);
}

// ==========================================================================

void MakeWindow(
   char * windowTitle,
   char * iconTitle,
   Pixmap iconPixmap,                      /* Pass (None) for none. */
   int minWidth, int minHeight,
   Boolean isOpenGLWindow,
   Pixel backgroundPixel,
   int argc, char *argv[]                  /* Command line arguments. */
) {
   XWMHints *wm_hints;                     /* Window manager hints. */
   XClassHint *class_hints;                /* Application class hints. */
   XSizeHints *size_hints;                 /* Preferred window geometry. */
   XTextProperty windowName,iconName;      /* Special X "strings" */
   int windowX, windowY;                   /* Upper left corner of window. */
   unsigned int windowBorderWidth;

   /* This stuff is for creating an OpenGL window. */
   GLXContext glxContext;
   XVisualInfo * visualInfo;
   XSetWindowAttributes setableAttributes;

   /*
      Allocate hint structures.
   */
   if (NULL == (wm_hints = XAllocWMHints()) ||
      NULL == (class_hints = XAllocClassHint()) ||
      NULL == (size_hints = XAllocSizeHints())
   ) {
      fprintf(stderr,"%s: failure allocating memory.\n", XStuff.commandLineName);
      exit(1);
   }

   /*
      Create text properties.
   */
   if (0 == XStringListToTextProperty(&windowTitle, 1, &windowName) ||
      0 == XStringListToTextProperty(&iconTitle, 1, &iconName)
   ) {
      fprintf(stderr,"%s: structure allocation for text property failed.\n",
         XStuff.commandLineName);
      exit(1);
   }

   if (isOpenGLWindow) {
      int attributeList[] = {
         GLX_RGBA,
         GLX_DOUBLEBUFFER,
         GLX_RED_SIZE, 1,
         GLX_GREEN_SIZE, 1,
         GLX_BLUE_SIZE, 1,
         GLX_DEPTH_SIZE, 1,
         None
      };

      /* Open an X visual */
      if ( bRootMode ) {

        XWindowAttributes xgwa;
        XVisualInfo vi_in, *vi_out;
        int out_count;

        XStuff.window = XStuff.rootWindow;
        XGetWindowAttributes (XStuff.display, XStuff.window, &xgwa);

        vi_in.screen = XStuff.screen_num;
        vi_in.visualid = XVisualIDFromVisual (xgwa.visual);
        vi_out = XGetVisualInfo (XStuff.display, VisualScreenMask|VisualIDMask,
                                 &vi_in, &out_count);
        if (!vi_out) {
          fprintf(stderr,"%s: can't get GL visual for window 0x%lx.\n",
                  XStuff.commandLineName, (unsigned long) XStuff.window);
          exit(1);
        }

        visualInfo = vi_out;
      } else {
        if (! (visualInfo = glXChooseVisual(XStuff.display, XStuff.screen_num, attributeList))) {
          fprintf(stderr,"%s: can't open GL visual.\n", XStuff.commandLineName);
          exit(1);
        }
      }

      /* Create a GLX context */
      if (!(glxContext = glXCreateContext(XStuff.display, visualInfo, None, GL_TRUE))) {
         fprintf(stderr,"%s: can't open GLX context.\n", XStuff.commandLineName);
         exit(1);
      }

      /* create a colormap */
      XStuff.colourMap = XCreateColormap(XStuff.display, XStuff.rootWindow, visualInfo->visual, AllocNone);
   }

   /*
      Create the main window.
   */
   windowX = XStuff.screenWidth/3;
   windowY = XStuff.screenHeight/3;
   XStuff.windowWidth = XStuff.screenWidth/3;
   XStuff.windowHeight = XStuff.screenHeight/3;
   windowBorderWidth = 0;

   if (isOpenGLWindow) {
      setableAttributes.colormap = XStuff.colourMap;
      setableAttributes.border_pixel = 0;

      if ( bRootMode ) {
         XStuff.window = XStuff.rootWindow;
      }
      else {
         XStuff.window = XCreateWindow(
            XStuff.display, XStuff.rootWindow,
            windowX, windowY, XStuff.windowWidth, XStuff.windowHeight,
            windowBorderWidth,
            visualInfo->depth, InputOutput, visualInfo->visual,
            CWBorderPixel | CWColormap,
            &setableAttributes);
      }
   }
   else {
      XStuff.window = XCreateSimpleWindow(XStuff.display, XStuff.rootWindow,
         windowX, windowY, XStuff.windowWidth, XStuff.windowHeight,
         windowBorderWidth,
         XStuff.blackPixel, XStuff.whitePixel);
   }

   /*
      Set window properties.
   */
   wm_hints->flags = StateHint | InputHint;
   wm_hints->initial_state = NormalState;   /* Should window be normal or iconified when first mapped ? */
   wm_hints->input = True;                  /* Does application need keyboard input? */
   if (None != iconPixmap) {
      wm_hints->flags |= IconPixmapHint;
      wm_hints->icon_pixmap = iconPixmap;
   }

   /* These are used by the window manager to get information */
   /* about this application from the resource database. */
   /*
      For example, with 4Dwm, you could add lines
      like these to your ~/.Xdefaults file :
              4Dwm*foo.clientDecoration: none
              4Dwm*goo.clientDecoration: border resizeh
      If foo is the name or class of an app's window, then that window will
      have no window decorations.  Similarly, in the case of goo, the window
      will only have a resizable border.  (The clientDecoration resource works
      the same way with Mwm.)
   */
   class_hints->res_name = XStuff.commandLineName;
   class_hints->res_class = "MyClassOfApplications";

   /* Use USPosition | USSize instead of PPosition | PSize to force hints. */
   /* size_hints->flags = PPosition | PSize | PMinSize; */
   size_hints->flags = PMinSize;
   /* In R4 and later, the x, y, width, and height members of XSizeHints should not be set. */
   size_hints->min_width = minWidth;
   size_hints->min_height = minHeight;

   if ( !bRootMode ) {
     XSetWMProperties(XStuff.display, XStuff.window, &windowName, &iconName, 
        argv, argc, size_hints, wm_hints,class_hints);
   }

   /*
      Select event types wanted.

      ConfigureNotify events inform the app that the window has been
      resized.  Processing ConfigureNotify events is more efficient
      than having to call XGetGeometry() (which requires a reply from
      the server) on every Expose event.
   */
   if ( ! bRootMode )
   XSelectInput(XStuff.display, XStuff.window,
      ExposureMask |
      KeyPressMask |
      KeyReleaseMask |
      ButtonPressMask |
      ButtonReleaseMask |
      PointerMotionMask |
      PointerMotionHintMask |
      StructureNotifyMask /* selects CirculateNotify, ConfigureNotify, DestroyNotify,
            GravityNotify, MapNotify, ReparentNotify, and UnmapNotify */
      );

   /*
      Display the window.
   */
   if ( ! bRootMode )
      XMapWindow(XStuff.display, XStuff.window);

   if (isOpenGLWindow) {
      /* Connect the context to the window. */
      if (!glXMakeCurrent(XStuff.display, XStuff.window, glxContext)) {
         fprintf(stderr,"%s: can't set GL context.\n", XStuff.commandLineName);
         exit(1);
      }
   }

   if (!isOpenGLWindow) {
      /*
         Set the background color.
      */
      XSetWindowBackground(XStuff.display, XStuff.window, backgroundPixel);
   }

   // If you don't like these fonts, use xfontsel
   // to find the names of ones you like better.
   //
   char * nameOfIdealBigFont = "-adobe-helvetica-bold-o-normal-*-14-*";
   char * nameOfIdealSmallFont = "-sgi-screen-medium-*-*-*-15-*-*-*-*-*-*-*";

   // These fonts will be used if the
   // former ones cannot be found.
   //
   char * nameOfHighlyPortableBigFont = "9x15";
   char * nameOfHighlyPortableSmallFont = "fixed";

   if (NULL == (XStuff.bigFont = XLoadQueryFont(XStuff.display,nameOfIdealBigFont)))
      if (NULL == (XStuff.bigFont = XLoadQueryFont(XStuff.display,nameOfHighlyPortableBigFont)))
         fprintf(stderr,"%s: can't open font \"%s\"\nSome text may not appear.\n", 
            XStuff.commandLineName, nameOfHighlyPortableBigFont
         );

   if (NULL == (XStuff.smallFont = XLoadQueryFont(XStuff.display,nameOfIdealSmallFont)))
      if (NULL == (XStuff.smallFont = XLoadQueryFont(XStuff.display,nameOfHighlyPortableSmallFont)))
         fprintf(stderr,"%s: can't open font \"%s\"\nSome text may not appear.\n",
            XStuff.commandLineName, nameOfHighlyPortableSmallFont
         );

   /*
      Get a graphics context.
   */
   XStuff.gc = XCreateGC(XStuff.display, XStuff.window, 0, NULL);
   /* Specify black foreground since default window background */
   /* is white and default foreground is undefined. */
   XSetForeground(XStuff.display, XStuff.gc, XStuff.blackPixel);

   /*
      Get geometry information about window
   */
   if (False == XGetGeometry(XStuff.display, XStuff.window, &XStuff.rootWindow,
      &windowX, &windowY, &XStuff.windowWidth, &XStuff.windowHeight,
      &windowBorderWidth, &XStuff.screenDepth)
   ) {
      fprintf(stderr,"%s: can't get window geometry.\n", XStuff.commandLineName);
      exit(1);
   }
}

// ==============================================================================

// The below bitmaps were created with the command "bitmap [filename]".

#define cursor_width 25
#define cursor_height 25
#define cursor_x_hot 12
#define cursor_y_hot 12
static char cursor_bits[] = {
   0xff, 0xff, 0xff, 0xff, 0xdf, 0xff, 0xf7, 0xff, 0xdf, 0xff, 0xf7, 0xff,
   0xaf, 0xff, 0xeb, 0xff, 0xaf, 0xff, 0xeb, 0xff, 0x77, 0xff, 0xdd, 0xff,
   0x77, 0xff, 0xdd, 0xff, 0xfb, 0xfe, 0xbe, 0xff, 0xfb, 0xfe, 0xbe, 0xff,
   0x01, 0x7c, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff,
   0xff, 0xd7, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff,
   0xff, 0xd7, 0xff, 0xff, 0xff, 0xd7, 0xff, 0xff, 0xff, 0xbb, 0xff, 0xff,
   0xff, 0xbb, 0xff, 0xff, 0xff, 0x7d, 0xff, 0xff, 0xff, 0x7d, 0xff, 0xff,
   0xff, 0x00, 0xfe, 0xff};

#define mask_width 25
#define mask_height 25
#define mask_x_hot 12
#define mask_y_hot 12
static char mask_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x08, 0x00, 0x20, 0x00, 0x08, 0x00,
   0x70, 0x00, 0x1c, 0x00, 0x70, 0x00, 0x1c, 0x00, 0xf8, 0x00, 0x3e, 0x00,
   0xf8, 0x00, 0x3e, 0x00, 0xfc, 0x01, 0x7f, 0x00, 0xfc, 0x01, 0x7f, 0x00,
   0xfe, 0x83, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00,
   0x00, 0x38, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00,
   0x00, 0x38, 0x00, 0x00, 0x00, 0x38, 0x00, 0x00, 0x00, 0x7c, 0x00, 0x00,
   0x00, 0x7c, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00,
   0x00, 0xff, 0x01, 0x00};

void CreateCursorForWindow() {

   Pixmap source, mask;
   XColor foreground, background;

   foreground.pixel = 0;
   foreground.red = 0;
   foreground.green = 65535;
   foreground.blue = 0;
   foreground.flags = DoRed | DoGreen | DoBlue;
   foreground.pad = 0;

   background.pixel = 0;
   background.red = 0;
   background.green = 0;
   background.blue = 0;
   background.flags = DoRed | DoGreen | DoBlue;
   background.pad = 0;

   source = XCreateBitmapFromData(XStuff.display, XStuff.window,
                   cursor_bits, cursor_width, cursor_height);
   mask = XCreateBitmapFromData(XStuff.display, XStuff.window,
                   mask_bits, mask_width, mask_height);
   XStuff.cursor = XCreatePixmapCursor(XStuff.display, source, mask,
                   &foreground, &background,
                   cursor_x_hot, cursor_y_hot);

   XDefineCursor(XStuff.display, XStuff.window, XStuff.cursor);

   XFreePixmap(XStuff.display, source);
   XFreePixmap(XStuff.display, mask);
}

// ==============================================================================

class Frustum {
public:
   void Init();
   void WindowResize();
   void Magnify(int delta);  // delta can be positive (to zoom in) or negative (to zoom out)
   void Reset();

   double Width() { return width; }
   double Height() { return height; }
   double LeftEdge() { return center_x - width/2; }
   double TopEdge() { return center_y + height/2; }
   double NearPlane() { return near; }
private:
   double unMagnifiedSize;
   double size;
   double width;   // width of near plane
   double height;  // height of near plane
   double near;    // distance to near plane from observer
   double far;     // distance to far plane from observer
   double center_x;
   double center_y;

   // magnificationFactor == magIncrement ^ magExponent
   //
   double magnificationFactor;  // 1.0 by default, larger values shrink the (width,height) proportionally
   int magExponent;             // 0 by default
   static const double magIncrement;
};

const double Frustum::magIncrement = 1.05;

void Frustum::Init() {

   near = objectDistanceFromCamera;
   far = near + 4*objectRadius;
   center_x = 0.0;
   center_y = 0.0;

   magnificationFactor = 1.0;
   magExponent = 0;

   // Create clipping volume.
   // Let r be objectRadius, d be objectDistanceFromCamera.
   // We have two similar right triangles
   //
   //                         |
   //                  |?     |r
   // eye  ____________|______|
   //            d        r
   //
   //
   //  ? = alpha = dr/(r+d)
   //
   // But this value of alpha makes the object unpleasantly
   // large, so we multiply by a constant greater than 1.0
   //
   double alpha = 1.25 * objectDistanceFromCamera * objectRadius / (objectRadius + objectDistanceFromCamera);
   unMagnifiedSize = 2*alpha;
   size = unMagnifiedSize / magnificationFactor;
   WindowResize();
}

void Frustum::WindowResize() {

   glViewport(0,0,XStuff.windowWidth,XStuff.windowHeight);
   if (XStuff.windowWidth < XStuff.windowHeight) {
      width = size;
      height = size * XStuff.windowHeight/(double)XStuff.windowWidth;
   }
   else {
      height = size;
      width = size * XStuff.windowWidth/(double)XStuff.windowHeight;
   }

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();   // clear matrix
   
   // for orthographic projection, use glOrtho() instead of glFrustum()
   //
   glFrustum(
      center_x - width/2, center_x + width/2, /* left,right */
      center_y - height/2, center_y + height/2, /* bottom, top */
      near, far /* near, far */
   );
}

void Frustum::Magnify(int delta) {

   int j;

   magExponent += delta;
   magnificationFactor = 1.0;
   if (magExponent < 0)
      for (j = magExponent; j < 0; ++j)
         magnificationFactor /= magIncrement;
   else for (j = 0; j < magExponent; ++j)
      magnificationFactor *= magIncrement;

   size = unMagnifiedSize / magnificationFactor;
   WindowResize();
}

void Frustum::Reset() {

   center_x = 0;
   center_y = 0;
   magExponent = 0;
   magnificationFactor = 1.0;
   size = unMagnifiedSize / magnificationFactor;
   WindowResize();
}

// ==============================================================================

Frustum frustum;

// ==============================================================================

class Sphere {
public:
    Sphere(double time = 0,int numStrips = 0,int numHemispheresToDisplay = 0,int numStripsToDisplay = 0,int numberOfLatitudinalPatchesPerHemisphere = 0,int numberOfLongitudinalPatchesPerStrip = 0) :
        matrixOfVertices(NULL) { Construct(time,numStrips,numHemispheresToDisplay,numStripsToDisplay,numberOfLatitudinalPatchesPerHemisphere,numberOfLongitudinalPatchesPerStrip); }
    ~Sphere();
    void Construct(double time = 0,int numStrips = 0,int numHemispheresToDisplay = 0,int numStripsToDisplay = 0,int numberOfLatitudinalPatchesPerHemisphere = 0,int numberOfLongitudinalPatchesPerStrip = 0);
    void Reconstruct() { DeallocateMatrix(); GenerateVertices(); }
    void Draw();
    void IncrementTime(double deltaTime) { if (Time < 1.0) { DeallocateMatrix(); Time+=deltaTime; if (deltaTime > 1.0) deltaTime = 1.0; GenerateVertices(); } }
    void DecrementTime(double deltaTime) { if (Time > 0.0) { DeallocateMatrix(); Time-=deltaTime; if (deltaTime < 0.0) deltaTime = 0.0; GenerateVertices(); } }
    void IncrementNumStrips() { DeallocateMatrix(); ++NumStrips; ++NumStripsToDisplay; GenerateVertices(); }
    void DecrementNumStrips() { if (NumStrips > 1) { DeallocateMatrix(); --NumStrips; if (NumStripsToDisplay > 1) --NumStripsToDisplay; GenerateVertices(); } }
    void ChangeNumHemispheresToDisplay() { if (NumHemispheresToDisplay == NumHemispheres) NumHemispheresToDisplay = 1; else ++NumHemispheresToDisplay; }
    void IncrementNumStripsToDisplay() { if (NumStripsToDisplay < NumStrips) {++NumStripsToDisplay;} }
    void DecrementNumStripsToDisplay() { if (NumStripsToDisplay > 1) {--NumStripsToDisplay;} }
    void IncrementLatitudinalResolution() { DeallocateMatrix(); ++NumberOfLatitudinalPatchesPerHemisphere; GenerateVertices(); }
    void DecrementLatitudinalResolution() { DeallocateMatrix(); --NumberOfLatitudinalPatchesPerHemisphere; GenerateVertices(); }
    void IncrementLongitudinalResolution() { DeallocateMatrix(); ++NumberOfLongitudinalPatchesPerStrip; GenerateVertices(); }
    void DecrementLongitudinalResolution() { DeallocateMatrix(); --NumberOfLongitudinalPatchesPerStrip; GenerateVertices(); }

    double GetTime() { return Time; }

private:
    void GenerateVertices();
    void DeallocateMatrix();

    double Time;   // between 0.0 and 1.0

    static const int NumHemispheres;
    int NumStrips;
    int NumHemispheresToDisplay;
    int NumStripsToDisplay;
    int NumberOfLatitudinalPatchesPerHemisphere,
        NumberOfLongitudinalPatchesPerStrip;

    // This matrix stores all the vertices and normals used to render the
    // sphere.  Elements in the matrix are arranged by [latitude][longitude].
    //
    GLPoint ** matrixOfVertices;
};

const int Sphere::NumHemispheres = 2;

Sphere::~Sphere() {

    DeallocateMatrix();
}

void Sphere::DeallocateMatrix() {

    int j;
    
    if (matrixOfVertices == NULL)
       return;
    for (j = NumberOfLatitudinalPatchesPerHemisphere; j >= 0; --j)
       delete [] (matrixOfVertices[j]);
    delete [] matrixOfVertices;
    matrixOfVertices = NULL;
}

void Sphere::Construct(
    double time,
    int numStrips,
    int numHemispheresToDisplay,
    int numStripsToDisplay,
    int numberOfLatitudinalPatchesPerHemisphere,
    int numberOfLongitudinalPatchesPerStrip
) {
    if (time < 0.0)
       Time = 0.0;
    else if (time > 1.0)
       Time = 1.0;
    else Time = time;

    NumStrips = numStrips < 1 ? defaultNumStrips : numStrips;

    NumHemispheresToDisplay = numHemispheresToDisplay < 1 ? NumHemispheres : numHemispheresToDisplay;

    NumStripsToDisplay = numStripsToDisplay < 1 ? NumStrips : numStripsToDisplay;

    if (numberOfLatitudinalPatchesPerHemisphere <= 0)
       NumberOfLatitudinalPatchesPerHemisphere = defaultNumberOfLatitudinalPatchesPerHemisphere;
    else
       NumberOfLatitudinalPatchesPerHemisphere = numberOfLatitudinalPatchesPerHemisphere;

    if (numberOfLongitudinalPatchesPerStrip <= 0)
       NumberOfLongitudinalPatchesPerStrip = defaultNumberOfLongitudinalPatchesPerStrip;
    else
       NumberOfLongitudinalPatchesPerStrip = numberOfLongitudinalPatchesPerStrip;

    GenerateVertices();
}

void Sphere::GenerateVertices() {

    int j;

    // clamp input parameters to their minima
    if (Time < 0.0)
       Time = 0.0;
    else if (Time > 1.0)
       Time = 1.0;

    if (NumberOfLatitudinalPatchesPerHemisphere < 2)
       NumberOfLatitudinalPatchesPerHemisphere = 2;

    if (NumberOfLongitudinalPatchesPerStrip < 2)
       NumberOfLongitudinalPatchesPerStrip = 2;

    // throw away previous geometry
    DeallocateMatrix();

    // allocate stuff
    matrixOfVertices = new GLPointPointer[1 + NumberOfLatitudinalPatchesPerHemisphere];
    for (j = NumberOfLatitudinalPatchesPerHemisphere; j >= 0; --j)
       matrixOfVertices[j] = new GLPoint[1 + NumberOfLongitudinalPatchesPerStrip];

    // generate the geometry
    generateGeometry(
       matrixOfVertices,
       Time,
       NumStrips,

       0.0,
       NumberOfLatitudinalPatchesPerHemisphere,
       1.0,
       0.0,
       NumberOfLongitudinalPatchesPerStrip,
       bHalfStrips ? 0.5 : 1.0
#ifdef BEND_IN /* this will display a cylindar bending into a sphere */
       ,Time
#endif
    );

}

void Sphere::Draw() {

    /*
    In a GL_TRIANGLE_STRIP block, if the vertices sent
    in are numbered 0,1,2,3,... the resulting strip will look like

        0---2---4---6---8--- ...
         \ / \ / \ / \ / \
          1---3---5---7---9 ...
 
    and will be cowposed of triangles (0,1,2), (2,1,3), (2,3,4), ...
    The ordering is important.  By convention, polygons whose vertices
    appear in counterclockwise order on the screen are front-facing.
    If you create polygons with vertices in the wrong order, and
    back-face culling is turned on, you'll never see the polygons
    appear on the screen.
    */

    int hemisphere,strip,j,k;

    glFrontFace(GL_CW);  // we're going to use the opposite convention

    glMatrixMode(GL_MODELVIEW);

    for (hemisphere = 0; hemisphere < NumHemispheresToDisplay; ++hemisphere) {
        glPushMatrix();
        glRotatef(hemisphere*180.0,0.0,1.0,0.0);
        for (strip = 0; strip < NumStripsToDisplay; ++strip) {
            glPushMatrix();
            glRotatef((hemisphere == 0 ? -strip : strip+1)*360.0/NumStrips,0.0,0.0,1.0);

            if (renderingStyle == style_points) {
                glBegin(GL_POINTS);
                for (j = 0; j <= NumberOfLatitudinalPatchesPerHemisphere; ++j)
                    for (k = 0; k <= NumberOfLongitudinalPatchesPerStrip; ++k) {
                        glNormal3fv(matrixOfVertices[j][k].normal);
                        glVertex3fv(matrixOfVertices[j][k].vertex);
                    }
                glEnd();
            }
            else {
                for (j = 0; j < NumberOfLatitudinalPatchesPerHemisphere; ++j) {
                    if (renderingStyle == style_polygons || (renderingStyle == style_bands && (j & 1)==hemisphere)) {
                        glBegin(GL_TRIANGLE_STRIP);
                        for (k = 0; k <= NumberOfLongitudinalPatchesPerStrip; ++k) {
                            glNormal3fv(matrixOfVertices[j][k].normal);
                            glVertex3fv(matrixOfVertices[j][k].vertex);
                            glNormal3fv(matrixOfVertices[j+1][k].normal);
                            glVertex3fv(matrixOfVertices[j+1][k].vertex);
                        }
                        glEnd();
                    }
                    else if (renderingStyle == style_checkered) {
                        for (k = j%2; k < NumberOfLongitudinalPatchesPerStrip; k+=2) {
                            glBegin(GL_TRIANGLE_STRIP);
                            glNormal3fv(matrixOfVertices[j][k].normal);
                            glVertex3fv(matrixOfVertices[j][k].vertex);
                            glNormal3fv(matrixOfVertices[j+1][k].normal);
                            glVertex3fv(matrixOfVertices[j+1][k].vertex);
                            glNormal3fv(matrixOfVertices[j][k+1].normal);
                            glVertex3fv(matrixOfVertices[j][k+1].vertex);
                            glNormal3fv(matrixOfVertices[j+1][k+1].normal);
                            glVertex3fv(matrixOfVertices[j+1][k+1].vertex);
                            glEnd();
                        }
                    }
                }
            }
            glPopMatrix();
        }
        glPopMatrix();
    }
}

// ==============================================================================

Sphere sphere;

// ==============================================================================

class RotationInformation {
public:
    RotationInformation()
        : centerX(0), centerY(0), viewportRadius(10),
          isClutchOn(False),
          previousDeviceX(0), previousDeviceY(0),
          rotationMode(arcball),
          elevation(0.0), azimuth(0.0)
    {
        CopyIdentityInto(oldRotationMatrix);
        CopyIdentityInto(rotationMatrix); 
    }

    void CalibrateDevice(int center_x, int center_y, int viewport_radius) {
        centerX = center_x;
        centerY = center_y;
        viewportRadius = viewport_radius;
    }

    Boolean IsClutchOn() { return isClutchOn; }
    void TurnClutchOn(int deviceX, int deviceY) {    // mouse button down
        isClutchOn = True;
        previousDeviceX = deviceX;
        previousDeviceY = deviceY;
    }
    void TurnClutchOff() {   // mouse button up
        isClutchOn = False;
    }
    void DeviceMotion(float deviceX, float deviceY);   // mouse motion

    enum RotationMode {   // CycleRotationMode() assumes that the first mode equals zero
        elevationAzimuth = 0,
        aroundXYAxes,
        virtualSphere,
        arcball,
        numberOfRotationModes
    };
    RotationMode GetRotationMode() { return rotationMode; }
    void SetRotationMode(RotationMode newMode);
    void CycleRotationMode() { SetRotationMode((RotationMode)((rotationMode+1) % (numberOfRotationModes))); }

    float * GetRotationMatrix() { return rotationMatrix; }
private:
    void CopyIdentityInto(float * matrix);

    int centerX, centerY, viewportRadius;
    Boolean isClutchOn;
    int previousDeviceX, previousDeviceY;
    RotationMode rotationMode;
    float oldRotationMatrix[16];   // cumulative past rotations
    float rotationMatrix[16];   // oldRotationMatrix + current rotation

    // stuff specific to "elevationAzimuth" mode
    float elevation, azimuth;   // in degrees; not radians
};

void RotationInformation::DeviceMotion(float deviceX, float deviceY) {

    // Here, the device axis conventions are X+ right, Y+ down
    // Note: glRotatef(theta, x, y, z); rotates theta degrees counter clockwise around vector (x,y,z)

    // The device was dragged deltaX pixels to the right and deltaY pixels down.
    float deltaX = deviceX - previousDeviceX,
          deltaY = deviceY - previousDeviceY;
    float pixelsPerDegree = 0.01*viewportRadius;
    
    if (!isClutchOn)
       return;

    glMatrixMode(GL_MODELVIEW);   // set transformation mode to model/view transformations
    glLoadIdentity();             // clear matrix

    switch (rotationMode) {
    case elevationAzimuth:
        elevation += deltaY/pixelsPerDegree;
        azimuth += deltaX/pixelsPerDegree;
        glRotatef(elevation,1.0,0.0,0.0);   // perform rotation around x+ axis (pointing right)
        glRotatef(azimuth,0.0,1.0,0.0);   // perform rotation around y+ axis (pointing up)
        glMultMatrixf(oldRotationMatrix);   // preserve previous rotations
        // don't save into old
        break;
    case aroundXYAxes:
        glRotatef(deltaY/pixelsPerDegree,1.0,0.0,0.0);   // perform rotation around x+ axis (pointing right)
        glRotatef(deltaX/pixelsPerDegree,0.0,1.0,0.0);   // perform rotation around y+ axis (pointing up)
        glMultMatrixf(oldRotationMatrix);   // preserve previous rotations
        glGetFloatv(GL_MODELVIEW_MATRIX, oldRotationMatrix);   // save result into old
        break;
    case virtualSphere:
    case arcball:
        float ux, uy, uz,   // vector u corresponding to previous mouse position, where |u| = 1
              uzSquared,
              vx, vy, vz,   // vector v corresponding to new mouse position, where |v| = 1
              wx, wy, wz,   // vector w is cross product (u x v)
              vzSquared;
        float theta,        // angle (in radians) between u and v
              dotProduct,
              shrinkFactor;

        // compute u
        //
        ux = (previousDeviceX - centerX) / (float)viewportRadius;
        uy = (centerY - previousDeviceY) / (float)viewportRadius;   // uy+ points up, not down
        uzSquared = 1.0 - ux*ux - uy*uy;
        if (uzSquared < 0.0) {
           // Shorten vector to force a magnitude of one
           shrinkFactor = sqrt(ux*ux + uy*uy);
           ux /= shrinkFactor;
           uy /= shrinkFactor;
           uz = 0.0;
        }
        else uz = sqrt(uzSquared);

        // compute v
        //
        vx = (deviceX - centerX) / (float)viewportRadius;
        vy = (centerY - deviceY) / (float)viewportRadius;   // vy+ points up, not down
        vzSquared = 1.0 - vx*vx - vy*vy;
        if (vzSquared < 0.0) {
           // Shorten vector to force a magnitude of one
           shrinkFactor = sqrt(vx*vx + vy*vy);
           vx /= shrinkFactor;
           vy /= shrinkFactor;
           vz = 0.0;
        }
        else vz = sqrt(vzSquared);

        // compute cross product w = (u x v)
        //
        wx =   uy*vz - vy*uz;
        wy = -(ux*vz - vx*uz);
        wz =   ux*vy - vx*uy;

        // Since
        //    |w| = |u||v||sin theta|
        // and
        //    |u| = |v| = 1
        // we have
        //    |w| = |sin theta|
        // But theta = asin(|w|) only if 0 <= theta <= pi/2 radians.
        // If theta > pi/2, theta != asin(|w|).
        //
        // So we calculate theta using the dot product (u.v)
        //    u.v = |u||v|(cos theta) = cos theta
        //    theta = acos(u.v)
        // This works for 0 <= theta <= pi radians.
        dotProduct = ux*vx + uy*vy + uz*vz;
        if (dotProduct < -1.0)
            dotProduct = -1.0;
        else if (dotProduct > 1.0)
            dotProduct = 1.0;
        theta = acos(dotProduct);
        if (rotationMode == arcball) {
            // This is the only difference between virtualSphere and arcball.
            // Note that this number could be changed to achieve a generalization
            // of virtualSphere/arcball ... although virtualSphere is the only mode
            // which does not exhibit hysterisis.
            theta *= 2.0;
        }
        glRotatef(180 * theta / PI, wx, wy, wz);
        glMultMatrixf(oldRotationMatrix);   // preserve previous rotations
        glGetFloatv(GL_MODELVIEW_MATRIX, oldRotationMatrix);   // save result into old
        break;
      default:
        abort();
        break;
    };
    
    glGetFloatv(GL_MODELVIEW_MATRIX, rotationMatrix);   // save result

    // Forgetting these two lines made for very interesting behaviour.
    // (Some bugs are cool !)  Try commenting them and see what happens ...
    previousDeviceX = (int)deviceX;
    previousDeviceY = (int)deviceY;
}

void RotationInformation::SetRotationMode(RotationMode newMode) {

    // Save rotations of current mode so they won't be lost,
    // allowing us to transfer smoothly into the new mode.
    memcpy(oldRotationMatrix,   // destination
           rotationMatrix,   // source
           16*sizeof(float));

    // reset mode-specific parameters
    elevation = 0.0;
    azimuth = 0.0;
    
    // set to new mode
    rotationMode = newMode;
}

void RotationInformation::CopyIdentityInto(float * matrix) {

    matrix[ 0] = 1.0;   matrix[ 1] = 0.0;   matrix[ 2] = 0.0;   matrix[ 3] = 0.0;
    matrix[ 4] = 0.0;   matrix[ 5] = 1.0;   matrix[ 6] = 0.0;   matrix[ 7] = 0.0;
    matrix[ 8] = 0.0;   matrix[ 9] = 0.0;   matrix[10] = 1.0;   matrix[11] = 0.0;
    matrix[12] = 0.0;   matrix[13] = 0.0;   matrix[14] = 0.0;   matrix[15] = 1.0;
}

// ==============================================================================

RotationInformation Rotator;

// ==============================================================================

void Refresh() {

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // clear image & z-buffer

   glMatrixMode(GL_MODELVIEW);   // set transformation mode to model/view transformations
   glLoadIdentity();             // clear matrix

   glTranslatef(0.0, 0.0, -(objectRadius + objectDistanceFromCamera));   // move object away from origin, out in front of camera
   glMultMatrixf(Rotator.GetRotationMatrix());   // perform rotations

   sphere.Draw();
   glFlush();
   glXSwapBuffers(XStuff.display, XStuff.window);

   if (bDisplayText) {

      // draw strings
      //
      char buffer[40];
      int x_offset = 5, y_offset = 0, charHeight;

      if (NULL != XStuff.smallFont) {
         XSetFont(XStuff.display, XStuff.gc, XStuff.smallFont->fid);
         charHeight = (int)(
            1.2 * ( XStuff.smallFont->ascent + XStuff.smallFont->descent )
         );

         strcpy(buffer,"rotation mode = ");
         switch(Rotator.GetRotationMode()) {
            case RotationInformation::elevationAzimuth : strcat(buffer,"elevation-azimuth"); break;
            case RotationInformation::aroundXYAxes     : strcat(buffer,"x-y");               break;
            case RotationInformation::virtualSphere    : strcat(buffer,"virtual sphere");    break;
            case RotationInformation::arcball          : strcat(buffer,"arcball");           break;
            default:
              abort();
              break;
         }
         XDrawString(XStuff.display, XStuff.window, XStuff.gc, x_offset, y_offset + charHeight, buffer, strlen(buffer));
         y_offset += charHeight;
      }

      if (NULL != XStuff.bigFont) {
         XSetFont(XStuff.display, XStuff.gc, XStuff.bigFont->fid);
         charHeight = (int)(
            1.2 * ( XStuff.bigFont->ascent + XStuff.bigFont->descent )
         );

         sprintf(buffer,"t = %.4f", sphere.GetTime());
         XDrawString(XStuff.display, XStuff.window, XStuff.gc, x_offset, y_offset + charHeight, buffer, strlen(buffer));
         y_offset += charHeight;
         sprintf(buffer,"dt = 1 / %d", (int)(1.0/deltaTime + 0.5));
         XDrawString(XStuff.display, XStuff.window, XStuff.gc, x_offset, y_offset + charHeight, buffer, strlen(buffer));
      }

      // If the user hits the key to advance time many times in rapid
      // succession, the cpu will start doing some heavy computations
      // to try and keep up, displaying frame after frame as time
      // advances.  Meanwhile, requests to the X server
      // to draw strings can queue up, and finally all get dumped
      // onto the last frame.  The user sees this as text
      // overwriting itself, becoming illegible.  By calling
      // XFlush(), we prevent the requests from queuing up.
      //
      XFlush(XStuff.display);
   }
}

// ==============================================================================

void MainLoop() {

   Window dummyWindow1, dummyWindow2;
   XEvent event;
   Boolean bQuit = False;
   int newMouseX, newMouseY,
       dummyInt1, dummyInt2,
       keyCode;
   unsigned int StateOfModifiers;
   
   while (!bQuit) {
      if (bAnimatedEversion || bAnimatedRotation) {
         while (0 == XEventsQueued(XStuff.display, QueuedAfterFlush)) {
            // Perform animation.
            //
            if (bAnimatedEversion) {
               if (bAnimatedEversionBackwards) {
                  if (sphere.GetTime() == 0.0)
                     bAnimatedEversionBackwards = ! bAnimatedEversionBackwards;
                  else
                     sphere.DecrementTime(deltaTime);
               }
               else {
                  if (sphere.GetTime() == 1.0)
                     bAnimatedEversionBackwards = ! bAnimatedEversionBackwards;
                  else
                     sphere.IncrementTime(deltaTime);
               }
            }
            if (bAnimatedRotation) {
               // Do not rotate if the user is
               // currently trying to rotate.
               //
               if (!Rotator.IsClutchOn()) {
                  double lengthOfMouseDrag;
                  lengthOfMouseDrag = XStuff.minOfWidthAndHeight()*0.5*speedOfAnimatedRotation;
                  Rotator.TurnClutchOn( XStuff.windowWidth/2, XStuff.windowHeight/2 );
                  Rotator.DeviceMotion(
                     XStuff.windowWidth/2  + lengthOfMouseDrag * cos(angleOfMouseDragForAnimatedRotation),
                     XStuff.windowHeight/2 + lengthOfMouseDrag * sin(angleOfMouseDragForAnimatedRotation)
                  );
                  Rotator.TurnClutchOff();
                  angleOfMouseDragForAnimatedRotation += 0.01;
                  if (angleOfMouseDragForAnimatedRotation > 2*PI)
                     angleOfMouseDragForAnimatedRotation -= 2*PI;
               }
            }
            Refresh();
         }
      }

      XNextEvent(XStuff.display, &event);
      switch (event.type) {
         case Expose:
            // Unless this is the last contiguous expose, don't redraw.
            if (event.xexpose.count != 0)
               break;
            // Only *contiguous* expose events can be skipped.  Searching through the event queue
            // to find the last expose event would skip over intervening ConfigureNotify events.

            Refresh();
            break;
         case ConfigureNotify:
            // window has been resized
            if (
               (int)XStuff.windowWidth != event.xconfigure.width ||
               (int)XStuff.windowHeight != event.xconfigure.height
            ) {
               XStuff.windowWidth = event.xconfigure.width;
               XStuff.windowHeight = event.xconfigure.height;

               frustum.WindowResize();
               Rotator.CalibrateDevice(XStuff.windowWidth/2, XStuff.windowHeight/2, XStuff.minOfWidthAndHeight()/2);
               Refresh();
            }
            break;
         case KeyPress:

            // Consult /usr/include/X11/keysymdef.h for keysym codes
            //
            switch (keyCode = XLookupKeysym(&event.xkey, 0)) {
               case XK_KP_Add:
               // case XK_plus:
                  sphere.IncrementTime(deltaTime);
                  Refresh();
                  break;
               case XK_KP_Subtract:
               // case XK_minus:
                  sphere.DecrementTime(deltaTime);
                  Refresh();
                  break;
               case XK_KP_Multiply:
                  deltaTime *= 2.0;
                  if (deltaTime > 1.0)
                     deltaTime = 1.0;
                  Refresh();
                  break;
               case XK_KP_Divide:
                  if (deltaTime * 0.5 >= 1.0/8192)
                     deltaTime *= 0.5;
                  Refresh();
                  break;
               case XK_Page_Up:
                  sphere.IncrementNumStrips();
                  Refresh();
                  break;
               case XK_Page_Down:
                  sphere.DecrementNumStrips();
                  Refresh();
                  break;
               case XK_Up:
                  frustum.Magnify(1);
                  Refresh();
                  break;
               case XK_Down:
                  frustum.Magnify(-1);
                  Refresh();
                  break;
               case XK_Home:
                  frustum.Reset();
                  Refresh();
                  break;
               case XK_Right:
                  sphere.IncrementNumStripsToDisplay();
                  Refresh();
                  break;
               case XK_Left:
                  sphere.DecrementNumStripsToDisplay();
                  Refresh();
                  break;
               case XK_End:
                  sphere.ChangeNumHemispheresToDisplay();
                  Refresh();
                  break;
               case XK_Delete:
                  bHalfStrips = !bHalfStrips;
                  sphere.Reconstruct();
                  Refresh();
                  break;
               case XK_F1:
                  sphere.IncrementLatitudinalResolution();
                  Refresh();
                  break;
               case XK_F2:
                  sphere.DecrementLatitudinalResolution();
                  Refresh();
                  break;
               case XK_F3:
                  sphere.IncrementLongitudinalResolution();
                  Refresh();
                  break;
               case XK_F4:
                  sphere.DecrementLongitudinalResolution();
                  Refresh();
                  break;
               case XK_F5:
                  bAnimatedEversion = !bAnimatedEversion;
                  break;
               case XK_F6:
                  bAnimatedRotation = !bAnimatedRotation;
                  break;
               case XK_F9:
                  bDisplayText = ! bDisplayText;
                  Refresh();
                  break;
#ifdef TOGGLEABLE_WINDOW_DECORATIONS
               case XK_F10:
                  PropMotifWmHints mwm_hints;
                  Atom mwm_hints_atom;

                  bNoWindowDecorations = ! bNoWindowDecorations;
                  mwm_hints_atom  = XInternAtom(
                     XStuff.display, _XA_MOTIF_WM_HINTS, False
                  );
                  mwm_hints.flags = (int) MWM_HINTS_DECORATIONS;
                  mwm_hints.decorations = bNoWindowDecorations ? 0
                     : MWM_DECOR_ALL;
                     //: MWM_DECOR_BORDER | MWM_DECOR_TITLE | MWM_DECOR_RESIZEH
                     //| MWM_DECOR_MINIMIZE | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU;
                  XChangeProperty (
                     XStuff.display, XStuff.window, mwm_hints_atom,
                     mwm_hints_atom,
                     32, PropModeReplace, (unsigned char *) & mwm_hints,
                     PROP_MWM_HINTS_ELEMENTS
                  );
                  break;
#endif
               case XK_space:
                  renderingStyle = (RenderingStyle)(((int)renderingStyle + 1) % number_of_styles);
                  Refresh();
                  break;
               case XK_S:
               case XK_s:
                  bShadingIsSmooth = !bShadingIsSmooth;
                  glShadeModel(bShadingIsSmooth ? GL_SMOOTH : GL_FLAT);
                  Refresh();
                  break;
               case XK_b:
               case XK_B:
                  // toggle back face culling
                  bBackFaceCulling = !bBackFaceCulling;
                  if (bBackFaceCulling) {
                     glEnable(GL_CULL_FACE);
                     glCullFace(bFrontFacing ? GL_BACK : GL_FRONT);
                  }
                  else
                     glDisable(GL_CULL_FACE);
                  Refresh();
                  break;
               case XK_f:
               case XK_F:
                  bFrontFacing = !bFrontFacing;
                  glCullFace(bFrontFacing ? GL_BACK : GL_FRONT);
                  Refresh();
                  break;
               case XK_r:
               case XK_R:
                  // cycle forward to new rotation mode
                  Rotator.CycleRotationMode();
                  Refresh();
                  break;
               case XK_1:
               case XK_2:
               case XK_3:
               case XK_4:
               case XK_5:
               case XK_6:
               case XK_7:
               case XK_8:
                  // change colour of something
                  int index;
                  GLfloat rgb[4];
                  index = keyCode - XK_1;
                  rgb[0] = (GLfloat)(index & 1);
                  rgb[1] = (GLfloat)((index >> 1) & 1);
                  rgb[2] = (GLfloat)((index >> 2) & 1);
                  rgb[3] = 1.0;

                  if ((event.xkey.state & Mod1Mask) && (event.xkey.state & ShiftMask)) {
                     // Alt-Shift
                     glClearColor(0.75*rgb[0], 0.75*rgb[1], 0.75*rgb[2], 0.0);
                  }
                  else if ((event.xkey.state & ControlMask) && (event.xkey.state & ShiftMask)) {
                     // Ctrl-Shift
                     glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, rgb);
                     rgb[0] = 1.0 - rgb[0];
                     rgb[1] = 1.0 - rgb[1];
                     rgb[2] = 1.0 - rgb[2];
                     glMaterialfv(GL_BACK, GL_AMBIENT_AND_DIFFUSE, rgb);
                  }
                  else if (event.xkey.state & ShiftMask) {
                     // Shift
                     glMaterialfv(GL_BACK, GL_AMBIENT_AND_DIFFUSE, rgb);
                  }
                  else if (event.xkey.state & Mod1Mask) {
                     // Alt
                     FreePixelColour(currentTextColour);
                     currentTextColour = AllocPixelColour(0.75*rgb[0], 0.75*rgb[1], 0.75*rgb[2]);
                     XSetForeground(XStuff.display, XStuff.gc, currentTextColour);
                  }
                  else {
                     // No modifiers
                     glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, rgb);
                  }
                  Refresh();
                  break;
               case XK_Escape:
                  // time to exit
                  bQuit = True;
                  break;
            }
            break;
         case KeyRelease:
            break;
         case ButtonRelease:
            Rotator.TurnClutchOff();
            break;
         case ButtonPress:
            XQueryPointer(XStuff.display, XStuff.window, &dummyWindow1,
                        &dummyWindow2, &dummyInt1, &dummyInt2,
                        &newMouseX, &newMouseY, &StateOfModifiers);
#ifndef TRY_SOMETHING_WEIRD
            Rotator.TurnClutchOn(newMouseX, newMouseY);
#else
            Rotator.TurnClutchOn(windowWidth/2, windowHeight/2);
            Rotator.DeviceMotion(newMouseX,newMouseY);
            Refresh();
#endif
            break;
         case MotionNotify:
            if (Rotator.IsClutchOn()) {
               XQueryPointer(XStuff.display, XStuff.window, &dummyWindow1,
                        &dummyWindow2, &dummyInt1, &dummyInt2,
                        &newMouseX, &newMouseY, &StateOfModifiers);
               Rotator.DeviceMotion(newMouseX,newMouseY);
               Refresh();
            }
            break;
      }   /* switch(event */
   }   /* while (!bQuit) */
}

// ==============================================================================

int main(int argc, char *argv[]) {

   char * display_name = NULL;       /* Server to connect to */

   XStuff.commandLineName = basename(argv[0]);

   if ( argc > 2 || (argc == 2 && strcmp(argv[1],"--root")!= 0)) {
      fprintf(stderr,"Usage: %s [--root]\n", argv[0] );
      exit(1);
   }
   if ( argc == 2 ) {
      bRootMode = True;
      bAnimatedEversion = True;
      bAnimatedRotation = True;
      bDisplayText = False;
   }

   // Connect to the X server.
   //
   if (NULL == (XStuff.display = XOpenDisplay(display_name))) {
      fprintf(stderr,"%s: can't connect to X server %s\n",
         XStuff.commandLineName, XDisplayName(display_name));
      exit(1);
   }
   XStuff.screen_num     = DefaultScreen         (XStuff.display);
   XStuff.screen_ptr     = DefaultScreenOfDisplay(XStuff.display);
   XStuff.screenWidth    = DisplayWidth          (XStuff.display, XStuff.screen_num);
   XStuff.screenHeight   = DisplayHeight         (XStuff.display, XStuff.screen_num);
   XStuff.screenWidthMM  = DisplayWidthMM        (XStuff.display, XStuff.screen_num);
   XStuff.screenHeightMM = DisplayHeightMM       (XStuff.display, XStuff.screen_num);
   XStuff.rootWindow     = RootWindow            (XStuff.display, XStuff.screen_num);
   XStuff.blackPixel     = BlackPixel            (XStuff.display, XStuff.screen_num);
   XStuff.whitePixel     = WhitePixel            (XStuff.display, XStuff.screen_num);

   if ( !bRootMode ) {
   printf(
      "This program displays a sphere undergoing the Thurston eversion.\n"
      "To rotate, click and drag with the left mouse button.\n"
      "Keys:\n"
      "  keypad \"+\"/\"-\"     Increase/decrease t by delta_t\n"
      "  keypad \"*\"/\"/\"     Multiply/divide delta_t by 2\n"
      "  page up/down       Increase/decrease total number of strips\n"
      "  up/down arrows     Zoom in/out\n"

      // Pan is not yet implemented, sorry.
      // "  home               Reset zoom and pan to initial values\n"
      "  home               Reset zoom to initial value\n"

      "  right/left arrows  Increase/decrease number of strips displayed\n"
      "  end                Toggle display of lower hemisphere\n"
      "  delete             Toggle display of half-strips\n"
      "  space              Cycle through polygons/checkered polygons/bands/points\n"
      "  F1/F2              Increase/decrease number of latitudinal patches\n"
      "  F3/F4              Increase/decrease number of longitudinal patches\n"
      "  F5/F6              Toggle animated eversion/rotation\n"
      "  F9                 Toggle text\n"
#ifdef TOGGLEABLE_WINDOW_DECORATIONS
      "  F10                Toggle window decorations\n"
      "                     (requires Motif-like window manager)\n"
#endif
      "  \"s\"                Toggle smooth/flat shading\n"
      "  \"b\"                Toggle backface culling\n"
      "  \"f\"                Toggle which faces are front facing\n"
      "  \"r\"                Cycle through rotation modes\n"
      "  \"1\"-\"8\"            Select colour of ...\n"
      "                       front faces (no modifiers), back faces (use Shift),\n"
      "                       front and back faces (use Ctrl-Shift),\n"
      "                       text (use Alt), background (use Alt-Shift)\n"
      "  escape             Quit\n"
      "Press Enter to begin ... "
   );
   getchar();
   }

   MakeWindow("Sphere Eversion","Eversion",None,0,0,True,XStuff.blackPixel,argc,argv);
   if ( !bRootMode ) CreateCursorForWindow();

   // Initialize some X stuff.
   //
   currentTextColour = AllocPixelColour(0.0, 0.75, 0.75);
   XSetForeground(XStuff.display, XStuff.gc, currentTextColour);

   // Initialize some OpenGL stuff.

   if (bBackFaceCulling) {
      // turn on back face culling
      glEnable(GL_CULL_FACE);
      glCullFace(bFrontFacing ? GL_BACK : GL_FRONT);
   }

   // Don't bother doing this, since the normals are
   // already normalized by us at generation time.
   // There's no need to ask OpenGL to explicitly try
   // to normalize again and again every time the
   // user rotates the view.
   //
   // glEnable(GL_NORMALIZE);   // This asks that normal vectors be normalized for us, automatically

   glClearColor(0.0, 0.0, 0.0, 0.0);   // this sets the colour to use when glClear() is called
   glClearDepth(1.0);
   // glDepthFunc(GL_LEQUAL);
   glEnable(GL_DEPTH_TEST);

   // Don't bother doing this
   //
   // // Clear the buffers.
   // //
   // glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   // glXSwapBuffers(display, window);
   // glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   // glXSwapBuffers(display, window);

   // Setup colours & lighting properties of the materiel.
   //
   GLfloat specular[] = { 1.0, 1.0, 1.0, 1.0 };
   GLfloat colourFront[] = { 0.0, 1.0, 1.0, 1.0 };   // colour of front faces of object
   GLfloat colourBack[] = { 1.0, 0.0, 0.0, 1.0 };    // colour of back faces of object
   GLfloat shininess[] = { 50.0 };
   glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
   glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, colourFront);
   glMaterialfv(GL_BACK, GL_AMBIENT_AND_DIFFUSE, colourBack);
   glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

   // Setup lighting.
   //
   GLfloat lightPosition[] = { 10.0, 10.0, 10.0, 0.0 }; 
   glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
   glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   GLfloat lightModelFlag[1] = { 1.0 };
   glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lightModelFlag);

   // Initialze some of my own stuff.
   //
   Rotator.CalibrateDevice(XStuff.windowWidth/2,XStuff.windowHeight/2,XStuff.minOfWidthAndHeight()/2);
   frustum.Init();

   // Loop for events until user quits.
   //
   MainLoop();

   // Clean up.
   //
   FreePixelColour(currentTextColour);
   if (NULL != XStuff.bigFont)
      XUnloadFont(XStuff.display, XStuff.bigFont->fid);
   if (NULL != XStuff.smallFont)
      XUnloadFont(XStuff.display, XStuff.smallFont->fid);
   XFreeGC(XStuff.display, XStuff.gc);
   XFreeCursor(XStuff.display, XStuff.cursor);
   XCloseDisplay(XStuff.display);
   return 0;
}

